import os
import datetime
import sqlite3

class Database(object):
    "A base class for database objects to inherit from."
    def __init__(self, file_path=None, db=None) -> None:
        """Initialize the database either by providing a file_path where a 
        new database will be created or provide the path to an existing db."""
        # set as current working directory if none provided
        file_path = file_path if file_path else os.getcwd()
        if db:
            self.connect(db)
        else:
            self.__create_db(file_path)
    
    def connect(self, db):
        "Create a connection to the db."
        self.cnx = sqlite3.connect(db)
        self.cur = self.cnx.cursor()

    def close(self):
        "Close the db connections."
        self.cnx.close()

    def __del__(self):
        self.close()

    def query(self, query, args=None):
        """Query the db with optional arguments. Query is executable and SQL
          statement and args should be a tuple."""
        try:
            if args:
                self.cur.execute(query, args)
            else:
                self.cur.execute(query)
        except sqlite3.Error as err: 
            print(err)

    def __create_db(self, file_path):
        pass

    def get_categories(self):
        pass


import os
import datetime
import sqlite3

from mybudgeter.database.database import Database

class Budget(Database):
    def __init__(self, file_path=None, db=None) -> None:
        """Initialize the database either by providing a file_path where a 
        new database will be created or provide the path to an existing db."""
        # set as current working directory if none provided
        file_path = file_path if file_path else os.getcwd()
        if db:
            self.connect(db)
        else:
            self.__create_db(file_path)

    def __create_db(self, file_path):
        "Create a new database and budget table to store budget information at the given file_path."
        db = os.path.join(file_path, "budgeting.db")
        query = """CREATE TABLE budget (
                            category text,
                            month int,
                            year int,
                            amount decimal(7,2),
                            PRIMARY KEY (category,month,year),
                            CHECK (month >= 1 AND month <= 12 AND year > 0)
                            );"""
        try:
            self.connect(db)
            self.query(query)
            self.cnx.commit()
        except sqlite3.Error as err: 
            print(err)
            self.close()

    def add_category(self, category:str, limit=0, date=None):
        """Add a new budget category to the budget table. 
        If no limit is given it will default to zero. 
        If no date is given the budget will be set for the current month and year.
        The date should be given as a datetime object."""

        date = datetime.datetime.now() if date is None else date
        args = (category.lower(), date.month, date.year, limit)
        query = """INSERT INTO budget 
                (category, month, year, amount) 
                VALUES (?, ?, ?, ?)"""
        self.query(query, args)
        self.cnx.commit()

    def add_from_lists(self, cat_list:list, limit_list=None, date_list=None):
        """Adds categories to the budget table from lists."""
        
        # check lists are same length and if not given, create default list
        if isinstance(limit_list, list):
            if not len(cat_list) == len(limit_list):
                raise Exception("The lists need to be the same length.")
        else:
            limit_list = [0]*len(cat_list)
            
        if isinstance(date_list, list):
            if not len(cat_list) == len(date_list):
                raise Exception("The lists need to be the same length.")
        else:
            date_list = [datetime.datetime.now()]*len(cat_list)
        
        # add to table
        for i in range(len(cat_list)):
            self.add_category(cat_list[i], limit_list[i], date_list[i])
    
    def modify_category(self, category:str, value:float, month:int, year:int):
        "Modifies the category limit for an entry in the budget table."

        query = f"update budget set amount = ? WHERE category = ? AND month = ? and year = ?"
        args = (value, category, month, year)
        self.query(query, args)
        self.cnx.commit()

    def delete_category(self, category:str, month:int, year:int):
        "Deletes a budget entry for the given category, month, year."

        query = f"delete from budget WHERE category = ? AND month = ? AND year = ?"
        args = (category, month, year)
        self.query(query, args)
        self.cnx.commit()

    def check_budget(self, month=None, year=None):
        """Returns a list of budget categories for the given month and year."""

        month = month if month else datetime.datetime.now().month
        year = year if year else datetime.datetime.now().year

        query = f"Select * from budget WHERE month = ? AND year = ?"
        args = (month, year)
        self.query(query, args)
        return [tup for tup in self.cur]
    
    def get_categories(self):
        "Returns a list of unique categories that are in the budget table."

        query = """SELECT DISTINCT category from budget"""
        self.query(query)
        return [cat[0] for cat in self.cur]


import unittest
from mybudgeter.tests.budget_test import TestBudget
from mybudgeter.tests.transactions_test import TestTransactions
from mybudgeter.tests.user_helper_test import TestUsers

def my_suite():
	suite = unittest.TestSuite()
	# add test for budget
	suite.addTest(unittest.makeSuite(TestBudget))
	# add test for transactions
	suite.addTest(unittest.makeSuite(TestTransactions))
	# add test for users and helper functions
	suite.addTest(unittest.makeSuite(TestUsers))

	runner = unittest.TextTestRunner() 
	print(runner.run(suite))

if __name__ == '__main__':
	my_suite()



import os
import matplotlib.pyplot as plt
from mybudgeter.database.budget import Budget
from mybudgeter.database.transactions import Transactions
from mybudgeter.utilities.helper import calculation_query, spending_query, linechart_query

# Create a user-defined exception to handle SQlite query that returns nothing
class SQliteError(Exception):
    def __init__(self, message) -> None:
        self.message = message

class User(object):
    TRANSACTION_TYPE = "transactions"
    BUDGET_TYPE = "budget"

    def __init__(self, file_path=None, db=None, budget=None, transactions=None) -> None:
        self.file_path = os.getcwd() if not file_path else file_path
        if db: # we assume the db has both tables
            self.budget = Budget(self.file_path, db=db)
            self.transactions = Transactions(self.file_path, db=db)
        else:
            # when budget is none, create new db, else can be a Budget object, or a path to budget db
            if isinstance(budget, Budget):
                self.budget = budget
            elif isinstance(budget, str):
                self.budget = Budget(self.file_path, budget)
            else:
                self.budget = Budget(self.file_path)
            
            if isinstance(transactions, Transactions):
                self.transactions = transactions
            elif isinstance(transactions, str):
                self.transactions = Transactions(self.file_path, transactions)
            else:
                self.transactions = Transactions(self.file_path)
    
    def total(self, type=TRANSACTION_TYPE, categories=None, months=None, years=None) -> float:
        """Calculate the total spending or budget."""
        # Build the base query
        try:
            base_query = f"SELECT SUM(amount) FROM {type}"
            where_clause, values = calculation_query(type, categories, months, years)
            full_query = f"{base_query}{' WHERE ' + where_clause if where_clause else ''}"

            # Execute the query
            getattr(self, type).query(query=full_query, args=values)
            total = getattr(self, type).cur.fetchone()[0]
            if not total:
                raise SQliteError("The SQlite query did not return anything, please check if you have entered the correct input parameter.")
            return total
        except (AttributeError, TypeError, SQliteError) as e:
            print(f"Error in querying the total: {e}")
            return None
        except Exception as e:
            print(f"Unexpected error in querying the total: {e}")
            raise 
    
    def average(self, type=TRANSACTION_TYPE, categories=None, months=None, years=None) -> float:
        """Calculate the average spending or budget."""
        # Build the base query
        try:
            base_query = f"SELECT AVG(amount) FROM {type}"
            where_clause, values = calculation_query(type, categories, months, years)
            full_query = f"{base_query}{' WHERE ' + where_clause if where_clause else ''}"

            # Execute the query
            getattr(self, type).query(query=full_query, args=values)
            average = getattr(self, type).cur.fetchone()[0]
            if not average:
                raise SQliteError("The SQlite query did not return anything, please check if you have entered the correct input parameter.")
            return average
        except (AttributeError, TypeError, SQliteError) as e:
            print(f"Error in querying the average: {e}")
            return None
        except Exception as e:
            print(f"Unexpected error in querying the average: {e}")
            raise 

    def remaining_budget(self, categories=None, months=None, years=None) -> float:
            """
            Calculate if the user is over or under budget base on the total in budget and spending.
            Default calculation will be the difference between the subtotal in budget and spending.
            user can choose to calculate the remaining budget for a given category in a specific month and year.
            warn the user if the user is overbudget
            """
            try:
                total_budget = self.total("budget", categories, months, years)
                total_transaction = self.total("transactions", categories, months, years)
                if total_budget != None and total_transaction != None:
                    remaining_budget = total_budget - total_transaction
                    if remaining_budget < 0:
                        print(f"Warning: You are over budget by ${abs(remaining_budget)}.")
                    return remaining_budget
                else:
                    raise SQliteError("The SQlite query did not return anything, please check if you have entered the correct input parameter.")
            except (AttributeError, TypeError, SQliteError) as e:
                print(f"Error in querying the remaining budget: {e}")
            except Exception as e:
                print(f"Unexpected error in querying the remaining budget: {e}")
                raise 

    def highest_spending(self, calculate_category=True):
            """
            Find the highest spending based on the provided input.
            - If calculate_category is True, find the highest spending category in general.
            - If calculate_category is False, find the highest spending month in the current year.
            """
            try:
                query, values = spending_query(calculate_category)
                self.transactions.query(query, values)
                result = self.transactions.cur.fetchone()
                if result:
                    query_type, amount = result
                    # print(f"Highest spending category is {query_type}: ${amount:.2f}") if calculate_category else print(f"Highest spending month in the current year is {query_type}: ${amount:.2f}")
                    return query_type, amount
                else:
                    raise SQliteError("There are no data found in the transactions database.")              
            except SQliteError as e:
                print(f"Error in querying the highest spending: {e}")
                return None
            except Exception as e:
                print(f"Unexpected error in querying the highest spending: {e}")
                raise

    def lowest_spending(self, calculate_category=True):
        """
        Find the lowest spending based on the provided input.
        - If calculate_category is True, find the lowest spending category in general.
        - If calculate_category is False, find the lowest spending month in the current year.
        """
        try:
            query, values = spending_query(calculate_category, "ASC")
            self.transactions.query(query, values)
            result = self.transactions.cur.fetchone()
            if result:
                query_type, amount = result
                # print(f"Lowest spending category is {query_type}: ${amount:.2f}") if calculate_category else print(f"Lowest spending month in the current year is {query_type}: ${amount:.2f}")
                return query_type, amount
            else:
                raise SQliteError("There are no data found in the tranasactions database.")          
        except SQliteError as e:
            print(f"Error in querying the lowest spending: {e}")
            return None
        except Exception as e:
            print(f"Unexpected error in querying the lowest spending: {e}")
            raise

    def summary(self):
        """
        Print out an overview of the user's budget information in a dashboard format.
        """
        remaining_budget = self.remaining_budget()
        highest_category, highest_cat_spending = self.highest_spending()
        lowest_category, lowest_cat_spending = self.lowest_spending()
        highest_month, highest_mo_spending = self.highest_spending(False)
        lowest_month, lowest_mo_spending = self.lowest_spending(False)

        # Print the dashboard
        print("\n===== Budget Summary =====")
        print(f"Remaining Budget: ${remaining_budget:.2f}\n")

        print("Category Breakdown:")
        print(f"Highest Spending Category: {highest_category.capitalize():<20}")
        print(f"Spending: ${highest_cat_spending:.2f}")
        print(f"Lowest Spending Category:{lowest_category.capitalize():<20}")
        print(f"Spending: ${lowest_cat_spending:.2f}\n")

        print("Monthly Breakdown:")
        print(f"Highest Spending Month: {highest_month.capitalize()}")
        print(f"Spending: ${highest_mo_spending:.2f}")
        print(f"Lowest Spending Month: {lowest_month.capitalize()}")
        print(f"Spending: ${lowest_mo_spending:.2f}\n")

        if remaining_budget < 0:
            print(f"Warning: Over budget by ${abs(remaining_budget):.2f}\n")
        else:
            print("You are within budget. Keep it up!\n")
        
    def pie_chart(self, type = TRANSACTION_TYPE):
        """
        Plot a pie chart displaying the budget or spending breakdown by categories.
        """
        query = f"SELECT category, SUM(amount) FROM {type} GROUP BY category"
        getattr(self, type).query(query)
        data = getattr(self, type).cur.fetchall()
        categories = [category[0] for category in data]
        amounts = [amount[1] for amount in data]

        # Create a pie chart
        plt.figure(figsize=(8, 8))
        plt.pie(amounts, labels=None, autopct='%1.1f%%', startangle=140)
        plt.title(f"{type.capitalize()} breakdown by categories")
        plt.legend(categories, title="Categories", loc="center left", bbox_to_anchor=(1, 0.5))
        plt.axis('equal')
        plt.show()

    def line_chart(self, type = TRANSACTION_TYPE):
        """
        Plot a line graph displaying the budget or spending trend.
        """
        getattr(self, type).query(linechart_query(type))
        data = getattr(self, type).cur.fetchall()
        months = [month[0] for month in data]
        amounts = [amount[1] for amount in data]

        # Plot the trend
        plt.figure(figsize=(10, 6))
        plt.plot(months, amounts, marker='o', linestyle='-')
        plt.title("Spending Trend Over the Last 12 Months" if type == "transactions" else f"{type.capitalize()} Trend Over the Last 12 Months")
        plt.xlabel("Month")
        plt.ylabel("Amount")
        plt.grid(True)
        plt.show()





import os
import datetime
import sqlite3

from mybudgeter.database.database import Database
        
class Transactions(Database):
    def __init__(self, file_path=None, db=None) -> None:
        """Initialize the database either by providing a file_path where a 
        new database will be created or provide the path to an existing db."""
        # set as current working directory if none provided
        file_path = file_path if file_path else os.getcwd()
        if db:
            self.connect(db)
        else:
            self.__create_db(file_path)

    def __create_db(self, file_path):
        db = os.path.join(file_path, "transactions.db")
        query = """CREATE TABLE transactions (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    trans_date date,
                    name text,
                    category text,
                    amount decimal(7,2))"""
        try:
            self.connect(db)
            self.query(query)
            self.cnx.commit()
        except sqlite3.Error as err: 
            print(err) 
            self.close()

    def add_transaction(self, category:str, amount:float, date=None, name=None):
        """Adds an individual transaction to the transactions table and returns the transaction id when successful.
        Date can be a datetime object or None which defaults to the current date. 
        Name is an optional string parameter to store info about the transaction more info."""
        
        if isinstance(date, datetime.datetime):
            # convert to string for sql
            date = date.strftime("%Y-%m-%d")
        elif date is None:
            date = datetime.datetime.now().strftime("%Y-%m-%d")
    
        args = (category.lower(), amount, date, name)

        # sql statement
        query = """INSERT INTO transactions (category, amount, trans_date, name) 
                                     VALUES (?, ?, ?, ?) RETURNING id"""
        self.query(query, args)
        res = self.cur.fetchone()[0]
        self.cnx.commit()
        return res

    def modify_transaction(self, id:int, field:str, value):
        """Modifies the value of the given field for the transaction in the table with id."""
        # change some part of the transaction
        query = f"update transactions set {field} = ? WHERE id = ?"
        args = (value, id)
        self.query(query, args)
        self.cnx.commit()

    def delete_transaction(self, id:int):
        "Deletes a transaction from the table."
        query = """delete from transactions WHERE id = ?"""
        args = (id,)
        self.query(query, args)
        self.cnx.commit()

    def get_categories(self):
        "Returns a list of unique categories that are in the transaction table."

        query = """SELECT DISTINCT category from transactions"""
        self.query(query)
        return [cat[0] for cat in self.cur]
    
    def get_n_transactions(self, n:int, sort_field='trans_date', asc=False):
        "Returns a list of n transactions sorted by the sort_field."
        if asc:
             query = f'SELECT * from transactions order by {sort_field} limit {n}'
        else:
            query = f'SELECT * from transactions order by {sort_field} DESC limit {n}'
        self.query(query)
        return [cat for cat in self.cur]


from datetime import datetime, timedelta

def calculation_query(type, categories=None, months=None, years=None):
    """Prepare conditions and values for the WHERE clause used in total and average in the User class"""
    conditions = []
    values = []

    if categories:
        if isinstance(categories, str):
            categories = [categories]
        placeholders = ', '.join('?' for _ in categories)
        conditions.append(f"category IN ({placeholders})")
        values.extend(categories)

    if months:
        if not isinstance(months, list):
            months = [months]
        if type == "transactions":
            month_placeholders = ', '.join('?' for _ in months)
            conditions.append(f"strftime('%m', trans_date) IN ({month_placeholders})")
            values.extend(str(month) for month in months)
        elif type == "budget":
            month_placeholders = ', '.join('?' for _ in months)
            conditions.append(f"month IN ({month_placeholders})")
            values.extend(months)

    if years:
        if not isinstance(years, list):
            years = [years]
        if type == "transactions":
            year_placeholders = ', '.join('?' for _ in years)
            conditions.append(f"strftime('%Y', trans_date) IN ({year_placeholders})")
            values.extend(str(year) for year in years)
        elif type == "budget":
            year_placeholders = ', '.join('?' for _ in years)
            conditions.append(f"year IN ({year_placeholders})")
            values.extend(years)
    # Add WHERE clause if conditions are present
    where_clause = " AND ".join(conditions)
    return where_clause, values

def spending_query(calculate_category=True, order_by="DESC"):
    """
    Find the spending based on the provided input.
    - If calculate_category is True, find the highest/lowest spending category in general.
    - If calculate_category is False, find the highest/lowest spending month in the current year.
    - order_by: "DESC" for highest spending, "ASC" for lowest spending
    """
    query_type = "category" if calculate_category else "strftime('%m', trans_date) as month"
    order = f"ORDER BY SUM(amount) {order_by}"

    current_year = datetime.now().year
    query = f"SELECT {query_type}, SUM(amount) FROM transactions"
    conditions = f" WHERE strftime('%Y', trans_date) = ?" if not calculate_category else ""
    query += f"{conditions} GROUP BY category {order} LIMIT 1" if calculate_category else f"{conditions} GROUP BY month {order} LIMIT 1"

    values = (str(current_year),) if not calculate_category else ()

    return query, values

def linechart_query(type):
    """Build query for the linechart function in the User class"""
    # Get the current date
    end_date = datetime.now()
    # Calculate the start date (12 months ago)
    start_date = end_date - timedelta(days=365)
    # Format the dates for the SQLite query
    start_date_str = start_date.strftime('%Y-%m-%d')
    end_date_str = end_date.strftime('%Y-%m-%d')
    if type == "transactions":
        query = f"SELECT strftime('%Y-%m', trans_date) AS month, SUM(amount) FROM {type} WHERE trans_date BETWEEN '{start_date_str}' AND '{end_date_str}' GROUP BY strftime('%Y-%m', trans_date) ORDER BY month"
    else:
        query = f"SELECT printf('%d-%02d', year, month) AS month, SUM(amount) FROM {type} WHERE printf('%d-%02d', year, month) BETWEEN '{start_date_str}' AND '{end_date_str}' GROUP BY year, month ORDER BY month"
    return query